# ----------------------------------------------------------------------------
#  CLOgmaNeo
#  Copyright(c) 2022 Ogma Intelligent Systems Corp. All rights reserved.
#
#  This copy of CLOgmaNeo is licensed to you under the terms described
#  in the CLOGMANEO_LICENSE.md file included in this distribution.
# ----------------------------------------------------------------------------

import pyopencl as cl
import importlib_resources as res

def load_prog(ctx: cl.Context, kernel_file: str = 'core.cl'):
    kernels_src = ''

    with res.open_text('ogmaneo', kernel_file) as f:
        kernels_src = f.read()

    return cl.Program(ctx, kernels_src)

def load_prog_extra(ctx: cl.Context):
    return load_prog(ctx, 'extra.cl')



# ----------------------------------------------------------------------------
#  CLOgmaNeo
#  Copyright(c) 2022 Ogma Intelligent Systems Corp. All rights reserved.
#
#  This copy of CLOgmaNeo is licensed to you under the terms described
#  in the CLOGMANEO_LICENSE.md file included in this distribution.
# ----------------------------------------------------------------------------

import numpy as np
import pyopencl as cl
import pyopencl.array
import pyopencl.clrandom
import math
from dataclasses import dataclass
import h5py
import pickle

class Encoder:
    @dataclass
    class VisibleLayerDesc:
        size: (int, int, int, int) # Width, height, column size, temporal size
        radius: int

    class VisibleLayer:
        weights: cl.array.Array
        reconstruction: cl.array.Array

    def __init__(self, cq: cl.CommandQueue, prog: cl.Program, hidden_size: (int, int, int) = (4, 4, 16), vlds: [ VisibleLayerDesc ] = [], grp: h5py.Group = None):
        if grp is None:
            self.hidden_size = hidden_size

            num_hidden_columns = hidden_size[0] * hidden_size[1]
            num_hidden_cells = num_hidden_columns * hidden_size[2]

            self.activations = cl.array.zeros(cq, (num_hidden_cells,), np.float32)
            self.hidden_states = cl.array.zeros(cq, (num_hidden_columns,), np.int32)

            self.vlds = vlds
            self.vls = []

            for i in range(len(vlds)):
                vld = self.vlds[i]
                vl = self.VisibleLayer()

                num_visible_columns = vld.size[0] * vld.size[1] * vld.size[3]
                num_visible_cells = num_visible_columns * vld.size[2]

                diam = vld.radius * 2 + 1
                area = diam * diam
                num_weights = num_hidden_cells * area * vld.size[2] * vld.size[3]

                vl.weights = cl.clrandom.rand(cq, (num_weights,), np.float32, a=-1.0, b=1.0)
                vl.reconstruction = cl.array.empty(cq, (num_visible_cells,), np.float32)

                vl.visible_states_prev = cl.array.zeros(cq, (num_visible_columns,), np.int32)

                self.vls.append(vl)

            self.history_pos_prev = 0

            # Hyperparameters
            self.lr = 0.1
            self.reg = 0.01

        else: # Load from h5py group
            self.hidden_size = pickle.loads(grp.attrs['hidden_size'].tobytes())
            
            num_hidden_columns = self.hidden_size[0] * self.hidden_size[1]
            num_hidden_cells = num_hidden_columns * self.hidden_size[2]

            self.activations = cl.array.empty(cq, (num_hidden_cells,), np.float32)
            self.hidden_states = cl.array.empty(cq, (num_hidden_columns,), np.int32)

            self.activations.set(np.array(grp['activations'][:], np.float32))
            self.hidden_states.set(np.array(grp['hidden_states'][:], np.int32))
            
            self.vlds = pickle.loads(grp.attrs['vlds'].tobytes())
            self.vls = []

            for i in range(len(self.vlds)):
                vld = self.vlds[i]
                vl = self.VisibleLayer()

                num_visible_columns = vld.size[0] * vld.size[1] * vld.size[3]
                num_visible_cells = num_visible_columns * vld.size[2]

                diam = vld.radius * 2 + 1
                area = diam * diam
                num_weights = num_hidden_cells * area * vld.size[2] * vld.size[3]

                vl.weights = cl.array.empty(cq, (num_weights,), np.float32)
                vl.reconstruction = cl.array.empty(cq, (num_visible_cells,), np.float32)

                vl.weights.set(np.array(grp['weights' + str(i)][:], np.float32))

                vl.visible_states_prev = cl.array.empty(cq, (num_visible_columns,), np.int32)
                vl.visible_states_prev.set(np.array(grp['visible_states_prev' + str(i)][:], np.int32))

                self.vls.append(vl)

            self.history_pos_prev = pickle.loads(grp.attrs['history_pos_prev'].tobytes())

            # Hyperparameters
            self.lr = pickle.loads(grp.attrs['lr'].tobytes())
            self.reg = pickle.loads(grp.attrs['reg'].tobytes())

        # Kernels
        self.accum_sparse_activations_kernel = prog.accum_sparse_activations
        self.sparse_activations_kernel = prog.sparse_activations
        self.dense_tanh_activations_kernel = prog.dense_tanh_activations
        self.encoder_learn_kernel = prog.encoder_learn

    def step(self, cq: cl.CommandQueue, visible_states: [ cl.array.Array ], errors: cl.array.Array, history_pos: int, learn_enabled: bool = True):
        assert(len(visible_states) == len(self.vls))

        # Pad 3-vecs to 4-vecs
        vec_hidden_size = np.array(list(self.hidden_size) + [ 1 ], dtype=np.int32)

        if learn_enabled:
            for i in range(len(self.vls)):
                vld = self.vlds[i]
                vl = self.vls[i]

                diam = vld.radius * 2 + 1

                vec_visible_size = np.array(list(vld.size), dtype=np.int32)

                self.encoder_learn_kernel(cq, self.hidden_size, (1, 1, self.hidden_size[2]),
                        vl.visible_states_prev.data, self.activations.data, errors.data, vl.weights.data, 
                        vec_visible_size, vec_hidden_size, np.int32(vld.radius),
                        np.int32(diam),
                        np.array([ vld.size[0] / self.hidden_size[0], vld.size[1] / self.hidden_size[1] ], dtype=np.float32),
                        np.int32(self.history_pos_prev),
                        np.float32(self.lr),
                        np.float32(self.reg))

        # Clear
        self.activations.fill(np.float32(0))

        # Accumulate for all visible layers
        for i in range(len(self.vls)):
            vld = self.vlds[i]
            vl = self.vls[i]

            diam = vld.radius * 2 + 1

            vec_visible_size = np.array(list(vld.size), dtype=np.int32)
            
            self.accum_sparse_activations_kernel(cq, self.hidden_size, (1, 1, self.hidden_size[2]),
                    visible_states[i].data, vl.weights.data, self.activations.data,
                    vec_visible_size, vec_hidden_size, np.int32(vld.radius), np.int32(diam),
                    np.array([ vld.size[0] / self.hidden_size[0], vld.size[1] / self.hidden_size[1] ], dtype=np.float32),
                    np.int32(history_pos))

        self.sparse_activations_kernel(cq, (self.hidden_size[0], self.hidden_size[1], 1), None, self.activations.data, self.hidden_states.data,
                vec_hidden_size,
                np.float32(1.0 / len(self.vls)))

        self.dense_tanh_activations_kernel(cq, self.hidden_size, None, self.activations.data,
                vec_hidden_size,
                np.float32(1.0)) # No scaling, as we did that in the previous activation step

        # Copy to prevs
        for i in range(len(self.vls)):
            vld = self.vlds[i]
            vl = self.vls[i]

            vl.visible_states_prev[:] = visible_states[i][:]

        self.history_pos_prev = history_pos

    def write(self, grp: h5py.Group):
        grp.attrs['hidden_size'] = np.void(pickle.dumps(self.hidden_size))

        grp.create_dataset('activations', data=self.activations.get())
        grp.create_dataset('hidden_states', data=self.hidden_states.get())

        grp.attrs['vlds'] = np.void(pickle.dumps(self.vlds))

        for i in range(len(self.vls)):
            grp.create_dataset('weights' + str(i), data=self.vls[i].weights.get())
            grp.create_dataset('visible_states_prev' + str(i), data=self.vls[i].visible_states_prev.get())

        grp.attrs['history_pos_prev'] = np.void(pickle.dumps(self.history_pos_prev))

        grp.attrs['lr'] = np.void(pickle.dumps(self.lr))
        grp.attrs['reg'] = np.void(pickle.dumps(self.reg))




# ----------------------------------------------------------------------------
#  CLOgmaNeo
#  Copyright(c) 2022 Ogma Intelligent Systems Corp. All rights reserved.
#
#  This copy of CLOgmaNeo is licensed to you under the terms described
#  in the CLOGMANEO_LICENSE.md file included in this distribution.
# ----------------------------------------------------------------------------

import numpy as np
import pyopencl as cl
import pyopencl.array
import pyopencl.clrandom
import math
from dataclasses import dataclass
from enum import Enum
from .encoder import Encoder
from .decoder import Decoder
import h5py
import pickle

class IOType(Enum):
    NONE = 0
    PREDICTION = 1

class Hierarchy:
    @dataclass
    class IODesc:
        size: (int, int, int) = (4, 4, 16)
        t: IOType = IOType.PREDICTION
        e_radius: int = 2
        d_radius: int = 2

    @dataclass
    class LayerDesc:
        hidden_size: (int, int, int) = (4, 4, 16)
        e_radius: int = 2
        d_radius: int = 2
        ticks_per_update: int = 2
        temporal_horizon: int = 4

    def __init__(self, cq: cl.CommandQueue, prog: cl.Program, io_descs: [ IODesc ] = [], lds: [ LayerDesc ] = [], grp: h5py.Group = None):
        if grp is None:
            self.io_descs = io_descs
            self.lds = lds

            self.encoders = []
            self.decoders = []
            self.histories = []
            self.errors = []
            self.slices = []

            # Create layers
            for i in range(len(lds)):
                e_vlds = []
                io_history = []

                if i == 0: # First layer
                    io_decoders = []

                    # For each IO layer
                    for j in range(len(io_descs)):
                        num_io_columns = io_descs[j].size[0] * io_descs[j].size[1]

                        # For each timestep
                        e_vlds.append(Encoder.VisibleLayerDesc(size=(io_descs[j].size[0], io_descs[j].size[1], io_descs[j].size[2], lds[i].temporal_horizon), radius=io_descs[j].e_radius))

                        io_history.append(cl.array.zeros(cq, (num_io_columns * lds[i].temporal_horizon,), np.int32))

                        if io_descs[j].t == IOType.PREDICTION:
                            if i < len(lds) - 1:
                                d_vlds = [ Decoder.VisibleLayerDesc(size=lds[i].hidden_size, radius=io_descs[j].d_radius, is_dense=True),
                                        Decoder.VisibleLayerDesc(size=lds[i].hidden_size, radius=io_descs[j].d_radius, is_dense=False) ]
                            else:
                                d_vlds = [ Decoder.VisibleLayerDesc(size=lds[i].hidden_size, radius=io_descs[j].d_radius, is_dense=True) ]

                            io_decoders.append(Decoder(cq, prog, (io_descs[j].size[0], io_descs[j].size[1], io_descs[j].size[2], 1), d_vlds))
                        else:
                            io_decoders.append(None) # Mark no decoder

                    self.decoders.append(io_decoders)

                else: # Higher layers
                    temporal_history = []

                    num_prev_columns = lds[i - 1].hidden_size[0] * lds[i - 1].hidden_size[1]

                    io_history.append(cl.array.zeros(cq, (num_prev_columns * lds[i].temporal_horizon,), np.int32))

                    e_vlds = [ Encoder.VisibleLayerDesc(size=(lds[i - 1].hidden_size[0], lds[i - 1].hidden_size[1], lds[i - 1].hidden_size[2], lds[i].temporal_horizon), radius=lds[i].e_radius) ]

                    if i < len(lds) - 1:
                        d_vlds = [ Decoder.VisibleLayerDesc(size=lds[i].hidden_size, radius=lds[i].d_radius, is_dense=True),
                                Decoder.VisibleLayerDesc(size=lds[i].hidden_size, radius=lds[i].d_radius, is_dense=False) ]
                    else:
                        d_vlds = [ Decoder.VisibleLayerDesc(size=lds[i].hidden_size, radius=lds[i].d_radius, is_dense=True) ]

                    temporal_decoders = [ Decoder(cq, prog, (lds[i - 1].hidden_size[0], lds[i - 1].hidden_size[1], lds[i - 1].hidden_size[2], lds[i].ticks_per_update), d_vlds) ]

                    self.decoders.append(temporal_decoders)

                self.encoders.append(Encoder(cq, prog, lds[i].hidden_size, e_vlds))

                self.histories.append(io_history)

                self.errors.append(cl.array.empty(cq, (self.lds[i].hidden_size[0] * self.lds[i].hidden_size[1] * self.lds[i].hidden_size[2],), dtype=np.float32))
                self.slices.append(cl.array.empty(cq, (self.lds[i].hidden_size[0] * self.lds[i].hidden_size[1],), dtype=np.int32))

            self.ticks = len(lds) * [ 0 ]
            self.ticks_per_update = [ lds[i].ticks_per_update for i in range(len(lds)) ]
            self.ticks_per_update[0] = 1 # First layer always 1

            self.history_pos = len(lds) * [ 0 ]

            self.updates = len(lds) * [ False ]

        else: # Load from h5py group
            self.io_descs = pickle.loads(grp.attrs['io_descs'].tobytes())
            self.lds = pickle.loads(grp.attrs['lds'].tobytes())

            self.encoders = []
            self.decoders = []
            self.histories = []
            self.errors = []
            self.slices = []

            # Create layers
            for i in range(len(self.lds)):
                io_history = []

                if i == 0: # First layer
                    io_decoders = []

                    # For each IO layer
                    for j in range(len(self.io_descs)):
                        num_io_columns = self.io_descs[j].size[0] * self.io_descs[j].size[1]

                        # For each timestep
                        io_history.append(cl.array.empty(cq, (num_io_columns * self.lds[i].temporal_horizon,), np.int32))
                        io_history[-1].set(np.array(grp['histories' + str(i) + '_' + str(j)][:], np.int32))

                        if self.io_descs[j].t == IOType.PREDICTION:
                            io_decoders.append(Decoder(cq, prog, grp=grp['decoders' + str(i) + '_' + str(j)]))
                        else:
                            io_decoders.append(None) # Mark no decoder

                    self.decoders.append(io_decoders)

                else: # Higher layers
                    num_prev_columns = self.lds[i - 1].hidden_size[0] * self.lds[i - 1].hidden_size[1]

                    io_history.append(cl.array.empty(cq, (num_prev_columns * self.lds[i].temporal_horizon,), np.int32))
                    io_history[-1].set(np.array(grp['histories' + str(i) + '_0'][:], np.int32))

                    temporal_decoders = [ Decoder(cq, prog, grp=grp['decoders' + str(i) + '_0']) ]

                    self.decoders.append(temporal_decoders)

                self.encoders.append(Encoder(cq, prog, grp=grp['encoders' + str(i)]))

                self.histories.append(io_history)

                self.errors.append(cl.array.empty(cq, (self.lds[i].hidden_size[0] * self.lds[i].hidden_size[1] * self.lds[i].hidden_size[2],), dtype=np.float32))
                self.slices.append(cl.array.empty(cq, (self.lds[i].hidden_size[0] * self.lds[i].hidden_size[1],), dtype=np.int32))

            self.ticks = pickle.loads(grp.attrs['ticks'].tobytes())
            self.ticks_per_update = pickle.loads(grp.attrs['ticks_per_update'].tobytes())

            self.history_pos = pickle.loads(grp.attrs['history_pos'].tobytes())

            self.updates = pickle.loads(grp.attrs['updates'].tobytes())

    def step(self, cq: cl.CommandQueue, input_states: [ cl.array.Array ], learn_enabled: bool = True):
        # Push front
        self.history_pos[0] -= 1

        if self.history_pos[0] < 0:
            self.history_pos[0] += self.lds[0].temporal_horizon

        # Push into first layer history
        for i in range(len(self.io_descs)):
            num_visible_columns = self.io_descs[i].size[0] * self.io_descs[i].size[1]

            self.histories[0][i][num_visible_columns * self.history_pos[0] : num_visible_columns * (self.history_pos[0] + 1)][:] = input_states[i][:]

        # Up-pass
        for i in range(len(self.encoders)):
            self.updates[i] = False

            if i == 0 or self.ticks[i] >= self.ticks_per_update[i]:
                self.ticks[i] = 0

                self.updates[i] = True

                encoder_visible_states = self.histories[i]

                if learn_enabled:
                    # Accumulate errors
                    self.errors[i].fill(np.float32(0))

                    if i == 0:
                        for j in range(len(self.io_descs)):
                            if self.decoders[i][j] is None:
                                continue

                            self.decoders[i][j].generate_errors(cq, 0, self.errors[i], input_states[j], 0, 1)
                    else:
                        self.decoders[i][0].generate_errors(cq, 0, self.errors[i], self.histories[i][0], self.history_pos[i], self.lds[i].temporal_horizon)

                self.encoders[i].step(cq, encoder_visible_states, self.errors[i], self.history_pos[i], learn_enabled)

                # If there is a higher layer
                if i < len(self.encoders) - 1:
                    i_next = i + 1

                    # Push front
                    self.history_pos[i_next] -= 1

                    if self.history_pos[i_next] < 0:
                        self.history_pos[i_next] += self.lds[i_next].temporal_horizon

                    num_visible_columns = self.lds[i].hidden_size[0] * self.lds[i].hidden_size[1]

                    self.histories[i_next][0][num_visible_columns * self.history_pos[i_next] : num_visible_columns * (self.history_pos[i_next] + 1)][:] = self.encoders[i].hidden_states[:]

                    self.ticks[i_next] += 1

        # Down-pass
        for i in range(len(self.decoders) - 1, -1, -1):
            if self.updates[i]:
                # Copy
                decoder_visible_states = []

                if i < len(self.lds) - 1:
                    num_hidden_columns = self.lds[i].hidden_size[0] * self.lds[i].hidden_size[1]
                    destride_index = self.ticks_per_update[i + 1] - 1 - self.ticks[i + 1]

                    self.slices[i][:] = self.decoders[i + 1][0].hidden_states[num_hidden_columns * destride_index : num_hidden_columns * (destride_index + 1)][:]

                    decoder_visible_states = [ self.encoders[i].activations, self.slices[i] ]
                else:
                    decoder_visible_states = [ self.encoders[i].activations ]

                if i == 0:
                    for j in range(len(self.io_descs)):
                        if self.decoders[i][j] is None:
                            continue

                        self.decoders[i][j].step(cq, decoder_visible_states, input_states[j], 0, 1, learn_enabled)
                else:
                    self.decoders[i][0].step(cq, decoder_visible_states, self.histories[i][0], self.history_pos[i], self.lds[i].temporal_horizon, learn_enabled)

    def get_predicted_states(self, i) -> cl.array.Array:
        assert(self.decoders[0][i] is not None)

        return self.decoders[0][i].hidden_states

    def write(self, grp: h5py.Group):
        grp.attrs['io_descs'] = np.void(pickle.dumps(self.io_descs))
        grp.attrs['lds'] = np.void(pickle.dumps(self.lds))

        for i in range(len(self.lds)):
            for j in range(len(self.histories[i])):
                grp.create_dataset('histories' + str(i) + '_' + str(j), data=self.histories[i][j].get())

            for j in range(len(self.decoders[i])):
                grp.create_group('decoders' + str(i) + '_' + str(j))
                self.decoders[i][j].write(grp['decoders' + str(i) + '_' + str(j)])

            grp.create_group('encoders' + str(i))
            self.encoders[i].write(grp['encoders' + str(i)])
            
        grp.attrs['ticks'] = np.void(pickle.dumps(self.ticks))
        grp.attrs['ticks_per_update'] = np.void(pickle.dumps(self.ticks_per_update))

        grp.attrs['history_pos'] = np.void(pickle.dumps(self.history_pos))

        grp.attrs['updates'] = np.void(pickle.dumps(self.updates))
